; PNG file functions
;
; Stan Hebben
;
; NOTE for masm forum users: this code is part of a larger project and won't
;   compile without modification, but feel free to base your work on this file.
;		This version only understands two formats, 24-bit RGB colors and 32-bit ARGB colors.
;

PNG_CHUNK_CRITICAL	equ	020000000h

.data
ZLIB_VERSION	db "1.2.3", 0

.code
ReadPNG proc inputfile:DWORD	; inputfile = pointer to raw png data

;	LOCAL inputfile:DWORD
	LOCAL _width:DWORD
	LOCAL _height:DWORD
	LOCAL bitdepth:DWORD
	LOCAL colortype:DWORD
	LOCAL filename:DWORD
	LOCAL palette:DWORD	; palette CHUNK
	LOCAL rawimagesize:DWORD
	LOCAL rawimage:DWORD
	LOCAL rawlinesize:DWORD
	LOCAL chunksize:DWORD
	LOCAL decompressing:DWORD
	LOCAL result:DWORD
	LOCAL bytesperpixel:DWORD
	LOCAL zstream:z_stream
	
	push esi
	push edi
	
	; Check file header
	cmp dword ptr [esi+0], 0474E5089h
	jnz _notvalidpng
	cmp dword ptr [esi+4], 00A1A0A0Dh
	jnz _notvalidpng
	
	; Check IHDR
	mov eax, [esi+8]
	bswap eax
	cmp eax, 13
	jnz _corrupt
	mov eax, [esi+12]
	bswap eax
	cmp eax, 'IHDR'
	jnz _corrupt
	
	; TODO: CRC check IHDR
	
	; Load png info
	mov eax, [esi+16]
	bswap eax
	mov _width, eax
	
	mov eax, [esi+20]
	bswap eax
	mov _height, eax
	
	movzx eax, byte ptr [esi+24]
	mov bitdepth, eax
	cmp eax, 17
	jnc _unknownbitdepth
	
	movzx eax, byte ptr [esi+25]
	mov colortype, eax
	cmp eax, 7
	jnc _unknowncolortype
	
	movzx eax, byte ptr [esi+26]
	cmp eax, 0
	jnz _unknowncompression
	
	movzx eax, byte ptr [esi+27]
	cmp eax, 0
	jnz _unknownfiltermethod
	
	movzx eax, byte ptr [esi+28]
	cmp eax, 2
	jnc _unknowninterlace
	
	add esi, 8+8+13+4
	
	; Allocate memory to store raw image data
	mov eax, colortype
	cmp eax, 0
	jz _alloc_c0
	cmp eax, 2
	jz _alloc_c2
	cmp eax, 3
	jz _alloc_c3
	cmp eax, 4
	jz _alloc_c4
	cmp eax, 6
	jz _alloc_c6
	jmp _unknowncolortype
	
_alloc_c0:
	mov ebx, bitdepth
	cmp ebx, 1
	jz _alloc_1bpp
	cmp ebx, 2
	jz _alloc_2bpp
	cmp ebx, 4
	jz _alloc_4bpp
	cmp ebx, 8
	jz _alloc_8bpp
	cmp ebx, 16
	jz _alloc_16bpp
	jmp _invalidcbcombo
	
_alloc_c2:
	mov ebx, bitdepth
	cmp ebx, 8
	jz _alloc_24bpp
	cmp ebx, 16
	jz _alloc_48bpp
	jmp _invalidcbcombo
	
_alloc_c3:
	mov ebx, bitdepth
	cmp ebx, 1
	jz _alloc_1bpp
	cmp ebx, 2
	jz _alloc_2bpp
	cmp ebx, 4
	jz _alloc_4bpp
	cmp ebx, 8
	jz _alloc_8bpp
	jmp _invalidcbcombo
	
_alloc_c4:
	mov ebx, bitdepth
	cmp ebx, 8
	jz _alloc_16bpp
	cmp ebx, 16
	jz _alloc_32bpp
	jmp _invalidcbcombo
	
_alloc_c6:
	mov ebx, bitdepth
	cmp ebx, 8
	jz _alloc_32bpp
	cmp ebx, 16
	jz _alloc_64bpp
	jmp _invalidcbcombo
	
_alloc_1bpp:
	mov bytesperpixel, 1
	mov eax, _width
	add eax, 15
	shr eax, 3	; (width + 15)/8 (1 extra byte -> filter)
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_2bpp:
	mov bytesperpixel, 1
	mov eax, _width
	add eax, 7
	shr eax, 2	; (width + 7)/8 (1 extra byte -> filter)
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_4bpp:
	mov bytesperpixel, 1
	mov eax, _width
	add eax, 3
	shr eax, 1	;	(width + 3)/2 (1 extra byte -> filter)
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_8bpp:
	mov bytesperpixel, 1
	mov eax, _width
	add eax, 1
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_16bpp:
	mov bytesperpixel, 2
	mov eax, _width
	lea eax, [2*eax+1]
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_24bpp:
	mov bytesperpixel, 3
	mov eax, _width
	lea eax, [2*eax+eax+1]
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_32bpp:
	mov bytesperpixel, 4
	mov eax, _width
	lea eax, [4*eax+1]
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_48bpp:
	mov bytesperpixel, 6
	mov eax, _width
	shl eax, 1
	lea eax, [2*eax+eax+1]
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_alloc_64bpp:
	mov bytesperpixel, 8
	mov eax, _width
	lea eax, [8*eax+1]
	mov rawlinesize, eax
	mul _height
	add eax, 4	; reserve bytes
	mov rawimagesize, eax
	jmp _startdecompress
	
_startdecompress:
	; Allocate
	push rawimagesize
	push LMEM_FIXED
	call LocalAlloc
	mov rawimage, eax
	
	; Now we can start decompressing all IDAT chunks in the file
	mov zstream.zalloc, 0
	mov zstream.zfree, 0
	mov zstream.opaque, 0
	
	mov zstream.next_in, 0
	mov zstream.avail_in, 0
	mov zstream.total_in, 0
	
	mov eax, rawimage
	mov zstream.next_out, eax
	mov eax, rawimagesize
	mov zstream.avail_out, eax
	mov zstream.total_out, 0
	
	push esi	; no idea if c calls preserve esi/edi
	
	push sizeof z_stream
	push offset ZLIB_VERSION
	lea eax, zstream
	push eax
	call inflateInit_
	add esp, 12
	
	cmp eax, Z_OK
	jnz _decompress_error
	
	mov decompressing, TRUE
	
	pop esi
	
_chunkloop:
	mov eax, [esi]
	
	bswap eax
	mov chunksize, eax
	mov eax, [esi+4]
	bswap eax
	cmp eax, 'PLTE'
	jnz @F
	mov palette, esi
	jmp _chunkskip
@@:
	cmp eax, 'IEND'
	jz _chunkexit
	cmp eax, 'IDAT'
	jz _chunkidat
	test eax, PNG_CHUNK_CRITICAL
	jz _unknowncriticalchunk
	jmp _chunkskip
	
_chunkidat:
	; TODO: CRC check data
	
	; decompress data
	lea eax, [esi+8]
	mov zstream.next_in, eax
	mov eax, chunksize
	mov zstream.avail_in, eax
	
	push esi;
	
	push Z_NO_FLUSH
	lea eax, zstream
	push eax
	call inflate
	add esp, 8
	
	pop esi;
	
	cmp eax, Z_OK
	jz _chunkskip
	
	cmp eax, Z_STREAM_END
	jz _chunkskip
	
	jmp _decompress_error
	
_chunkskip:
	add esi, 12
	add esi, chunksize
	jmp _chunkloop
	
_chunkexit:
	lea eax, zstream
	push eax
	call inflateEnd
	add esp, 4
	
	mov decompressing, FALSE
	
	; Unfilter everything
	mov edx, 0
	mov esi, rawimage
_unfilter_loop:
	push edx
	
	movzx eax, byte ptr [esi]
	cmp eax, 0
	jz _unfilter_skip	; no filter
	
	cmp eax, 1
	jnz @F
	
	; sub (working)
	mov ebx, bytesperpixel
	mov edi, esi
	add edi, rawlinesize
	lea esi, [esi+ebx+1]
	neg ebx
_unfilter_sub:
	mov eax, [esi]			; current
	add eax, [esi+ebx]	; previous
	mov [esi], al
	add esi, 1
	cmp esi, edi
	jnz _unfilter_sub
	jmp _unfilter_cont
	
@@:
	cmp eax, 2
	jnz @F
	
	cmp edx, 0
	jz _unfilter_skip	; don't filter
	
	; up
	mov ebx, rawlinesize
	mov edi, esi
	add edi, ebx
	add esi, 1
	neg ebx
_unfilter_up:
	mov eax, [esi]			; current
	add eax, [esi+ebx]; above
	mov [esi], al
	add esi, 1
	cmp esi, edi
	jnz _unfilter_up
	jmp _unfilter_cont
	
@@:
	cmp eax, 3
	jnz @F
	
	; TODO: Handle first line differently
	
	; average
	mov ebx, bytesperpixel
	mov ecx, rawlinesize
	lea edx, [esi+ebx+1]
	lea edi, [esi+ecx]
	add esi, 1
	
	neg ebx
	neg ecx
	
	; first few bytes
_unfilter_avg1:
	movzx eax, byte ptr [esi+ecx]	; esi[*-linelength]
	shr eax, 1
	add eax, [esi]
	mov [esi], al
	add esi, 1
	cmp esi, edx
	jnz _unfilter_avg1
	
_unfilter_avg2:
	movzx eax, byte ptr[esi+ebx]
	movzx edx, byte ptr[esi+ecx]
	add eax, edx
	shr eax, 1
	add eax, [esi]
	mov [esi], al
	add esi, 1
	cmp esi, edi
	jnz _unfilter_avg2
	jmp _unfilter_cont
	
@@:
	cmp eax, 4
	jnz _corrupt
	
	; paeth
	
	cmp edx, 0
	jnz @F
	
	jmp _unfilter_skip
	; TODO: Handle first line differently
@@:
	
	mov ecx, rawlinesize
	lea edi, [esi+ecx]
	push edi
	neg ecx
	
	mov eax, bytesperpixel
	lea edi, [esi+eax+1]
	add esi, 1
_unfilter_paeth1:
	mov eax, [esi+ecx]
	add eax, [esi]
	mov [esi], al
	
	add esi, 1
	cmp esi, edi
	jnz _unfilter_paeth1
	
	pop edi
	
	mov ebx, bytesperpixel
	neg ebx
	mov edx, ebx
	add edx, ecx
	
_unfilter_paeth2:
	push ebx
	push ecx
	push edx
	
	movzx eax, byte ptr [esi+ebx]
	movzx ebx, byte ptr [esi+ecx]
	movzx ecx, byte ptr [esi+edx]
	
	mov edx, eax
	add edx, ebx
	sub edx, ecx
	
	sub eax, edx
	jge @F
	neg eax
@@:
	sub ebx, edx
	jge @F
	neg ebx
@@:
	sub ecx, edx
	jge @F
	neg ecx
@@:
	
	cmp ebx, eax
	jc _unfilter_paeth_bc
	cmp ecx, eax
	jc _unfilter_paeth_bc
	
	pop edx
	pop ecx
	pop ebx
	mov eax, [esi+ebx]
	jmp _unfilter_paeth_pred
	
_unfilter_paeth_bc:
	cmp ecx, ebx
	jc _unfilter_paeth_c
	
	pop edx
	pop ecx
	pop ebx
	mov eax, [esi+ecx]
	jmp _unfilter_paeth_pred
	
_unfilter_paeth_c:
	pop edx
	pop ecx
	pop ebx
	mov eax, [esi+edx]
	jmp _unfilter_paeth_pred
	
_unfilter_paeth_pred:
	add eax, [esi]
	mov [esi], al
	
	add esi, 1
	cmp esi, edi
	jnz _unfilter_paeth2
	
	jmp _unfilter_cont
	
_unfilter_skip:
	add esi, rawlinesize
_unfilter_cont:
	pop edx
	add edx, 1
	cmp edx, _height
	jnz _unfilter_loop
_unfilter_done:
	; Final step: convert to Image
	
	mov eax, _width
	mul _height
	lea ecx, [Image_headersize + 4*eax]
	push ecx
	push LMEM_FIXED
	call LocalAlloc
	mov result, eax
	
	mov ebx, _width
	mov [eax+Image_width], ebx
	mov ebx, _height
	mov [eax+Image_height], ebx
	
	mov esi, rawimage
	mov edi, result
	add edi, Image_headersize
	
	mov eax, colortype
	cmp eax, 0
	jz _convert_c0
	cmp eax, 2
	jz _convert_c2
	cmp eax, 3
	jz _convert_c3
	cmp eax, 4
	jz _convert_c4
	cmp eax, 6
	jz _convert_c6
	jmp _done	; no report needed, already checked for this before
	
_convert_c0:
	mov eax, bitdepth
	cmp eax, 1
	jz _convert_c0b1
	cmp eax, 2
	jz _convert_c0b2
	cmp eax, 4
	jz _convert_c0b4
	cmp eax, 8
	jz _convert_c0b8
	cmp eax, 16
	jz _convert_c0b16
	jmp _done
	
_convert_c2:
	mov eax, bitdepth
	cmp eax, 8
	jz _convert_c2b8
	cmp eax, 16
	jz _convert_c2b16
	jmp _done
	
_convert_c3:
	mov eax, bitdepth
	cmp eax, 1
	jz _convert_c3b1
	cmp eax, 2
	jz _convert_c3b2
	cmp eax, 4
	jz _convert_c3b4
	cmp eax, 8
	jz _convert_c3b8
	jmp _done
	
_convert_c4:
	mov eax, bitdepth
	cmp eax, 8
	jz _convert_c4b8
	cmp eax, 16
	jz _convert_c4b16
	jmp _done
	
_convert_c6:
	mov eax, bitdepth
	cmp eax, 8
	jz _convert_c6b8
	cmp eax, 16
	jz _convert_c6b16
	jmp _done

; TODO: Add more conversions
_convert_c0b1:	; 1 bpp black/white
_convert_c0b2:	; 2 bpp greyscale
_convert_c0b4:	; 4 bpp greyscale
_convert_c0b8:	; 8 bpp greyscale
_convert_c0b16:	; 16 bpp greyscale
;	debugpt 10
	jmp _done

_convert_c2b8:	; 24 bpp RGB
	mov ecx, _width
	mov edx, _height
_convert_c2b8_loop1:
	add esi, 1
_convert_c2b8_loop2:
	mov eax, [esi]
	bswap eax
	shr eax, 8
	or eax, 0FF000000h
	mov [edi], eax
	add edi, 4
	add esi, 3
	sub ecx, 1
	jnz _convert_c2b8_loop2
	mov ecx, _width
	sub edx, 1
	jnz _convert_c2b8_loop1
	jmp _success

_convert_c2b16:	; 48 bpp RGB
_convert_c3b1:	; 2-color palette
_convert_c3b2:	; 4-color palette
_convert_c3b4:	; 16-color palette
_convert_c3b8:	; 256-color palette
_convert_c4b8:	; 16 bpp greyscale/alpha
_convert_c4b16:	; 32 bpp greyscale/alpha
;	debugpt 10
	jmp _done
	
_convert_c6b8:	;	32 bpp ARGB
	mov ecx, _width
	mov edx, _height
_convert_c6b8_loop1:
	add esi, 1
_convert_c6b8_loop2:
	mov eax, [esi]
	mov ebx, eax
	and ebx, 0FF000000h
	bswap eax
	shr eax, 8
	or eax, ebx
	mov [edi], eax
	add esi, 4
	add edi, 4
	sub ecx, 1
	jnz _convert_c6b8_loop2
	mov ecx, _width
	sub edx, 1
	jnz _convert_c6b8_loop1
	jmp _success

_convert_c6b16:	; 64 bpp ARGB
;	debugpt 10
	jmp _done

_success:
	; free structures
	mov eax, inputfile
	cmp eax, NULL
	jz @F
	push eax
	call LocalFree
@@:
	mov eax, rawimage
	cmp eax, NULL
	jz @F
	push eax
	call LocalFree
@@:
	
	; restore state
	pop edi
	pop esi
	mov eax, result
	ret

;_notvalidpng:
;	push filename
;	push STRID_PNG_NOTAPNGFILE
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 12
;	
;	jmp _done
;
;_corrupt:
;	push filename
;	push STRID_PNG_CORRUPT
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 12
;	
;	jmp _done
;
;_unknownbitdepth:
;	push eax	; bit depth value
;	push filename
;	push STRID_PNG_UNKNOWN_BITDEPTH
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done
;
;_unknowncolortype:
;	push eax	; color type value
;	push filename
;	push STRID_PNG_UNKNOWN_COLORTYPE
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done
;
;_unknowncompression:
;	push eax	; compression value
;	push filename
;	push STRID_PNG_UNKNOWN_COMPRESSION
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done
;
;_unknownfiltermethod:
;	push eax	; bit depth value
;	push filename
;	push STRID_PNG_UNKNOWN_FILTERMETHOD
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done
;
;_unknowninterlace:
;	push eax	; bit depth value
;	push filename
;	push STRID_PNG_UNKNOWN_INTERLACEMETHOD
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done
;
;_invalidcbcombo:
;	push ebx	; bitdepth
;	push eax	; colortype
;	push filename
;	push STRID_PNG_INVALIDCBCOMBO
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 20
;	
;	jmp _done
;
;_unknowncriticalchunk:
;	push eax	; chunk
;	push filename
;	push STRID_PNG_UNKNOWNCRITICALCHUNK
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done
;
;_decompress_error:
;	push eax	; error code
;	push filename
;	push STRID_PNG_DECOMPRESSERROR
;	call GetLangString
;	push eax
;	push offset LastError
;	call wsprintfW
;	add esp, 16
;	
;	jmp _done

_done:	; only do this in case of an error
	; free structures
	mov eax, filename
	cmp eax, NULL
	jz @F
	push eax
	call LocalFree
@@:
	mov eax, inputfile
	cmp eax, NULL
	jz @F
	push eax
	call LocalFree
@@:
	mov eax, rawimage
	cmp eax, NULL
	jz @F
	push eax
	call LocalFree
@@:
	cmp decompressing, FALSE
	jz @F
	lea eax, zstream
	push eax
	call inflateEnd
	add esp, 4
@@:
	cmp result, NULL
	jz @F
	push result
	call LocalFree
@@:
	
	; restore state
	pop edi
	pop esi
	mov eax, 0
	ret

ReadPNG endp
